/*
 * Copyright 2013 Alexander Baumgartner
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package at.jku.risc.stout.urauc;

import java.awt.Container;
import java.awt.Dimension;
import java.awt.Frame;
import java.io.IOException;
import java.io.OutputStream;
import java.io.PrintStream;
import java.text.DecimalFormat;
import java.text.NumberFormat;
import java.util.Set;

import javax.swing.JApplet;
import javax.swing.JScrollPane;
import javax.swing.JTextArea;
import javax.swing.SwingUtilities;
import javax.swing.UIManager;
import javax.swing.text.BadLocationException;

import at.jku.risc.stout.urauc.algo.AlignFnc;
import at.jku.risc.stout.urauc.algo.DebugLevel;
import at.jku.risc.stout.urauc.data.InputParser;
import at.jku.risc.stout.urauc.data.atom.TermAtom;
import at.jku.risc.stout.urauc.util.ControlledException;

public class AntiUnifyApplet extends JApplet {
	private static final long serialVersionUID = -1261469987435837893L;

	private JTextArea textArea = new JTextArea();
	private JScrollPane scroll = new JScrollPane(textArea,
			JScrollPane.VERTICAL_SCROLLBAR_AS_NEEDED,
			JScrollPane.HORIZONTAL_SCROLLBAR_AS_NEEDED);

	private static final int MAX_LINES = 5000;
	// Jitter for MAX_LINES = 4900 - 5100
	private static final int MAX_LINES_VARIANZ = 100;
	private static final NumberFormat LINENUM_FORMAT = new DecimalFormat(
			"000000: ");
	private static final String PRINT_HEAD = "";
	private static final String PRINT_TAIL = "DONE";

	private final PrintStream out = new PrintStream(new OutputStream() {
		StringBuilder buffer = new StringBuilder();
		long count = 0L;

		@Override
		public void write(int b) throws IOException {
			buffer.appendCodePoint(b);
			if (b == '\n') {
				count++;
				flush();
			}
		}

		@Override
		public void flush() throws IOException {
			display(LINENUM_FORMAT.format(count) + buffer.toString());
			buffer.setLength(0);
		}
	});
	private final PrintStream info = new PrintStream(new OutputStream() {
		StringBuilder buffer = new StringBuilder();

		@Override
		public void write(int b) throws IOException {
			buffer.appendCodePoint(b);
			if (b == '\n')
				flush();
		}

		@Override
		public void flush() throws IOException {
			display("INFO:  " + buffer.toString());
			buffer.setLength(0);
		}
	});
	private final PrintStream err = new PrintStream(new OutputStream() {
		StringBuilder buffer = new StringBuilder();

		@Override
		public void write(int b) throws IOException {
			buffer.appendCodePoint(b);
			if (b == '\n')
				flush();
		}

		@Override
		public void flush() throws IOException {
			display("ERROR: " + buffer.toString());
			buffer.setLength(0);
		}
	});

	@Override
	public void init() {
		textArea.setEditable(false);
		scroll.setPreferredSize(new Dimension(getWidth(), getHeight() - 5));
		add(scroll);

		try {
			UIManager.setLookAndFeel(UIManager.getSystemLookAndFeelClassName());
			SwingUtilities.updateComponentTreeUI(this);
		} catch (Exception isOk_UseDefaultLookAndFeel) {
		}
		Container c = this;
		while (c.getParent() != null) {
			c = c.getParent();
			if (c instanceof Frame)
				((Frame) c).setResizable(true);
		}
		setVisible(true);
		validate();
	}

	@Override
	public void resize(int width, int height) {
		super.resize(width, height);
		scroll.setPreferredSize(new Dimension(width, height - 5));
	}

	@Override
	public void start() {
		new Thread() {
			@Override
			public synchronized void start() {
				setDaemon(true);
				super.start();
			}

			@Override
			public void run() {
				display(PRINT_HEAD);
				try {
					String aupsIn = getParameter("problemset");
					if (aupsIn == null || aupsIn.length() == 0) {
						err.println("No anti-unification problem sent by the browser.\n\n"
								+ "This might be a communication-error between your browser and the Java-Applet.\n"
								+ "Please try it once more. If you use Opera, please turn off the Turbo mode.\n"
								+ "If the problem persists, please try another browser.");
					} else {
						/*******************************
						 * Parse input
						 *******************************/
						final DebugLevel debugLevel = DebugLevel
								.valueOf(getParameter("debugmode"));
						boolean iterateAll = "true"
								.equals(getParameter("iterateall"));
						boolean justify = "true"
								.equals(getParameter("justify"));
						boolean applyResH = "true"
								.equals(getParameter("applyresh"));
						final boolean equalityFilter = "true"
								.equals(getParameter("equalityfilter"));
						boolean commRoot = "true"
								.equals(getParameter("commutativeroot"));
						Set<TermAtom> commFnc = InputParser
								.parseAtomSet(getParameter("commutativesymbols"));
						Set<TermAtom> assocFnc = InputParser
								.parseAtomSet(getParameter("associativesymbols"));
						AlignFnc aFnc = (AlignFnc) Class.forName(
								getParameter("alignfnc")).newInstance();
						int minLen = 0;
						try {
							minLen = Integer.parseInt(getParameter("minlen"));
						} catch (Exception e) {
						}
						int upperBound = 0;
						try {
							upperBound = Integer
									.parseInt(getParameter("upperbound"));
						} catch (Exception e) {
						}

						info.println("Alignment computation function = "
								+ getParameter("alignfnc"));
						info.println("Justify computed generalization = "
								+ justify);
						info.println("Iterate all possibilities = "
								+ iterateAll);
						if (!"true"
								.equals(getParameter("ignorealgebraicaxioms"))) {
							info.println("Minimum alignment length = " + minLen);
							info.println("Upper bound for commutative arrangement iteration = "
									+ upperBound);
							info.println("Apply equality filter = "
									+ equalityFilter);
						}
						info.println("Output format = " + debugLevel);
						info.println();
						if (!"true"
								.equals(getParameter("ignorealgebraicaxioms"))) {
							info.println("Commutative root hedge = " + commRoot);
							info.println("Commutative symbols = " + commFnc);
							info.println("Associative symbols = " + assocFnc);
							info.println();
						}
						info.println("PROBLEM:");

						/*******************************
						 * Call the algorithm
						 *******************************/
						new AntiUnifyConsole(debugLevel, info, out).exec(
								aupsIn, getParameter("alignment"), iterateAll,
								justify, applyResH, equalityFilter, commRoot,
								commFnc, assocFnc, aFnc, minLen, upperBound);
					}
				} catch (ControlledException ex) {
					err.println(ex.getMessage() + " ("
							+ ex.getClass().getSimpleName() + ")");
				} catch (Throwable ex) {
					ex.printStackTrace(err);
				}
				display(PRINT_TAIL);
			}
		}.start();
	}

	private void display(String str) {
		try {
			if (textArea.getLineCount() >= MAX_LINES + MAX_LINES_VARIANZ) {
				textArea.replaceRange(null, 0,
						textArea.getLineEndOffset(MAX_LINES_VARIANZ * 2));
			}
			textArea.append(str);
			repaint();
		} catch (BadLocationException e) {
			e.printStackTrace();
		}
	}

	/**
	 * May be called by JavaScript to enable copying of text from the applet to
	 * the system clipboard.
	 */
	public String getSelectedText() {
		String text = textArea.getSelectedText();
		return text == null ? "" : text;
	}
}
