/*
 * Copyright 2013 Alexander Baumgartner
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package at.jku.risc.stout.urauc;

import java.io.PrintStream;
import java.util.Arrays;
import java.util.Collections;
import java.util.Set;

import at.jku.risc.stout.urauc.algo.AlignFnc;
import at.jku.risc.stout.urauc.algo.DebugLevel;
import at.jku.risc.stout.urauc.data.InputParser;
import at.jku.risc.stout.urauc.data.atom.TermAtom;

public class AntiUnifyMain {

	static DebugLevel debugLevel = DebugLevel.SIMPLE;
	static boolean iterateAll = false;
	static boolean commRoot = false;
	static boolean justify = false;
	static boolean equalityFilter = false;
	static boolean applyResH = false;
	static Set<TermAtom> commFnc = Collections.emptySet();
	static Set<TermAtom> assocFnc = Collections.emptySet();
	static int minLen = 0;
	static int upperBound = 0;
	static String aFncIn = "at.jku.risc.stout.urauc.algo.AlignFncLAA";
	static String alignIn = "";

	public static void main(String[] args) {
		if (args.length >= 1) {
			/*******************************
			 * Parse input
			 *******************************/
			if ("-d".equals(args[0])) {
				debugLevel = DebugLevel.PROGRESS;
				main(Arrays.copyOfRange(args, 1, args.length));
			} else if ("-v".equals(args[0])) {
				if (debugLevel != DebugLevel.PROGRESS)
					debugLevel = DebugLevel.VERBOSE;
				main(Arrays.copyOfRange(args, 1, args.length));
			} else if ("-a".equals(args[0])) {
				iterateAll = true;
				main(Arrays.copyOfRange(args, 1, args.length));
			} else if ("-h".equals(args[0])) {
				printHelp(System.out);
			} else if ("-j".equals(args[0])) {
				justify = true;
				main(Arrays.copyOfRange(args, 1, args.length));
			} else if ("-p".equals(args[0])) {
				applyResH = true;
				main(Arrays.copyOfRange(args, 1, args.length));
			} else if ("-e".equals(args[0])) {
				equalityFilter = true;
				main(Arrays.copyOfRange(args, 1, args.length));
			} else if ("-r".equals(args[0])) {
				commRoot = true;
				main(Arrays.copyOfRange(args, 1, args.length));
			} else if ("-c".equals(args[0])) {
				try {
					commFnc = InputParser.parseAtomSet(argumentPair(args));
				} catch (Exception e) {
				}
			} else if ("-l".equals(args[0])) {
				alignIn = argumentPair(args);
			} else if ("-s".equals(args[0])) {
				try {
					assocFnc = InputParser.parseAtomSet(argumentPair(args));
				} catch (Exception e) {
				}
			} else if ("-m".equals(args[0])) {
				try {
					minLen = Integer.parseInt(argumentPair(args));
				} catch (Exception e) {
				}
			} else if ("-u".equals(args[0])) {
				try {
					upperBound = Integer.parseInt(argumentPair(args));
				} catch (Exception e) {
				}
			} else if ("-f".equals(args[0])) {
				aFncIn = argumentPair(args);
			} else {
				try {
					/*******************************
					 * Call the algorithm
					 *******************************/
					new AntiUnifyConsole(debugLevel, System.out, System.out)
							.exec(args[0], alignIn, iterateAll, justify,
									applyResH, equalityFilter, commRoot,
									commFnc, assocFnc, (AlignFnc) Class
											.forName(aFncIn).newInstance(),
									minLen, upperBound);

					System.out.println();
					System.out.println("Supported by the Austrian Science Fund"
							+ " (FWF) under the project SToUT P 24087-N18");
				} catch (Exception e) {
					e.printStackTrace();
				}
			}
		} else {
			argumentPair(args); // print error message
		}
	}

	private static String argumentPair(String[] args) {
		if (args.length < 2) {
			System.out.println("Insufficient parameters!");
			System.out.println();
			printHelp(System.out);
			return null;
		}
		main(Arrays.copyOfRange(args, 2, args.length));
		return args[1];
	}

	private static void printHelp(PrintStream out) {
		System.out.println("Anti-unification problem solver for unranked "
				+ "terms and hedges");
		System.out.println();
		System.out.println("Input example: \"(f(b, g(a)), a, f(h(a, z))) =^= "
				+ "f(b, g(a, f(h(a, x))))\"");
		System.out.println();
		System.out.println("Simple execution example: "
				+ "java at.jku.risc.stout.urauc.AntiUnifyMain -a "
				+ "\"f(a,b) =^= f(a,b,c)\"");
		System.out.println("Advanced execution example with jvm parameters:");
		System.out.println("java -Xmx2048m -XX:MaxPermSize=1024m -Xss128m "
				+ "at.jku.risc.stout.urauc.AntiUnifyMain -a -j -c \"f\" -f "
				+ "\"at.jku.risc.stout.urauc.algo.AlignFncLAA\" "
				+ "\"f(b,a) =^= f(a,b,c)\"");
		System.out.println();
		System.out.println("  -j    justify computed generalization");
		System.out.println("  -a    iterate all possibilities");
		System.out.println("  -p    apply Res-H rule");
		System.out.println("  -f X  set alignment computation function to "
				+ "X where X is the full class name");
		System.out.println("  -l X  set an alignment X where X is of the "
				+ "form f<1.2, 2.2.1>g<2, 3>...");
		System.out.println("  -r    make root hedge commutative");
		System.out.println("  -c X  make the given function symbols X "
				+ "commutative where X is a comma separated list");
		System.out.println("  -s X  make the given function symbols X "
				+ "associative where X is a comma separated list");
		System.out.println("  -m X  set minimum alignment length "
				+ "X where X is a positive integer");
		System.out.println("  -u X  set upper bound X for commutative "
				+ "arrangement iteration where X is a positive integer");
		System.out.println("  -e    enable commutative equality filter");
		System.out.println("  -v    verbose output mode prints sigma "
				+ "and the store");
		System.out.println("  -d    debug mode prints out every available "
				+ "information");
		System.out.println("  -h    print this information");
	}
}
