/*
 * Copyright 2013 Alexander Baumgartner
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package at.jku.risc.stout.urauc.algo;

import java.io.PrintStream;

import at.jku.risc.stout.urauc.algo.AlignFnc.AlignmentIterator;
import at.jku.risc.stout.urauc.algo.AntiUnifyProblem.VariableWithHedges;
import at.jku.risc.stout.urauc.data.Alignment;
import at.jku.risc.stout.urauc.data.EquationSystem;
import at.jku.risc.stout.urauc.data.Hedge;
import at.jku.risc.stout.urauc.data.NodeFactory;
import at.jku.risc.stout.urauc.data.atom.HedgeVar;
import at.jku.risc.stout.urauc.data.atom.Variable;
import at.jku.risc.stout.urauc.util.ControlledException;

/**
 * This class encapsulates the rule based system {@linkplain AntiUnifySystem}
 * and allows to justify the result by back substitution of the generalization
 * with the reported differences inside the computed store.
 * 
 * @author Alexander Baumgartner
 */
public class AntiUnify {
	private AlignFnc aFnc;
	private EquationSystem<AntiUnifyProblem> sys;
	private int minAlignmentLength;
	private int upperBound;
	private DebugLevel debugLevel;
	private boolean applyResH;

	public AntiUnify(AlignFnc aFnc, EquationSystem<AntiUnifyProblem> sys,
			DebugLevel debugLevel) {
		this(aFnc, sys, 0, -1, false, debugLevel);
	}

	public AntiUnify(AlignFnc aFnc, EquationSystem<AntiUnifyProblem> sys,
			int minAlignmentLength, int upperBound, boolean applyResH,
			DebugLevel debugLevel) {
		this.aFnc = aFnc;
		this.sys = sys;
		this.minAlignmentLength = minAlignmentLength;
		this.upperBound = upperBound;
		this.applyResH = applyResH;
		this.debugLevel = debugLevel;
		AlignFnc.debugLevel = debugLevel;
	}

	public long antiUnify(boolean iterateAll, boolean justify, PrintStream out)
			throws ControlledException, JustificationException {
		long count = 0;
		for (AntiUnifyProblem aup : sys)
			count += antiUnifySingleAup(aup, iterateAll, justify, out);
		return count;
	}

	/** TODO: Override and instantiate the extended algorithm. */
	protected AntiUnifySystem createSystem(
			EquationSystem<AntiUnifyProblem> eqSys) {
		return new AntiUnifySystem(eqSys, new Substitution(), applyResH);
	}

	public long antiUnifySingleAup(AntiUnifyProblem aup, boolean iterateAll,
			boolean justify, PrintStream out) throws ControlledException,
			JustificationException {
		if (debugLevel == DebugLevel.PROGRESS)
			out.println("Computing commutative arrangements:\nThis may take some time...");
		long start = System.currentTimeMillis();
		MaximumIterationException thrownEx = null;
		AlignmentIterator iter;
		try {
			iter = aFnc.getIterator(aup, minAlignmentLength, upperBound,
					iterateAll, out);
		} catch (MaximumIterationException e) {
			iter = e.getPartialIterator();
			thrownEx = e;
		}
		if (debugLevel == DebugLevel.PROGRESS) {
			out.println("Computation of commutative arrangements done. Time needed: "
					+ (System.currentTimeMillis() - start) / 1000.0);
			out.println();
		}

		long count = 0;
		while (iter.hasNext()) {
			NodeFactory.resetCounter();
			aup.setAlignment(iter.next());
			AntiUnifyOne rau = new AntiUnifyOne(aup);
			rau.antiUnify(out);
			if (justify)
				rau.justifyGeneralization(out);
			callback(rau.getResult(), rau.getMostGeneral());
			if (!iterateAll)
				return 1;
			count++;
		}
		if (debugLevel != DebugLevel.SILENT)
			out.println(count + " generalizations with alignment length "
					+ iter.getMaxAlignmentLen() + " found.");
		if (thrownEx != null)
			throw thrownEx;
		return count;
	}

	/**
	 * This callback function will be invoked for every found generalization. By
	 * default it does nothing. For instance, one can use this callback function
	 * to collect all the results or to filter out some interesting results.<br>
	 * In combination with a matching algorithm, one can obtain the minimal
	 * complete set of generalizations by using this callback function.
	 */
	public void callback(AntiUnifySystem result, Variable generalizationVar) {
	}

	public class AntiUnifyOne {
		private Hedge leftHedge, rightHedge;
		private Alignment alignment;
		private HedgeVar mostGeneral;
		private AntiUnifySystem uraucSys;

		public AntiUnifyOne(AntiUnifyProblem aup) throws ControlledException {
			this.leftHedge = aup.getLeft();
			this.rightHedge = aup.getRight();
			this.alignment = aup.getAlignment();
		}

		/**
		 * Calls the rule based system {@linkplain AntiUnifySystem} and tries to
		 * justify the computed result if the second argument of this method is
		 * true.
		 */
		public void antiUnify(PrintStream out) throws ControlledException {
			EquationSystem<AntiUnifyProblem> eqSys = new EquationSystem<AntiUnifyProblem>() {
				@Override
				public AntiUnifyProblem newEquation() {
					return new AntiUnifyProblem();
				}
			};
			AntiUnifyProblem aupClone = new AntiUnifyProblem(leftHedge.clone(),
					rightHedge.clone(), alignment.clone());
			eqSys.add(aupClone);
			mostGeneral = aupClone.getHorizontalVarX();
			uraucSys = createSystem(eqSys);
			uraucSys.compute(debugLevel, out);
			debug(out);
		}

		private void debug(PrintStream out) {
			if (debugLevel == DebugLevel.SIMPLE) {
				out.println(" Result: " + uraucSys);
			} else if (debugLevel != DebugLevel.SILENT) {
				out.println("-----------");
				out.println("   Result: " + uraucSys);
				out.println("-----------");
				out.println("Alignment: " + alignment);
				out.println("    Store: " + uraucSys.storeToString());
				out.println();
			}
		}

		public void justifyGeneralization(PrintStream out)
				throws JustificationException {
			Hedge generalization = uraucSys.getSigma().get(mostGeneral);
			Hedge leftInstance = generalization.clone();
			Hedge rightInstance = generalization.clone();
			if (out != null && debugLevel == DebugLevel.PROGRESS)
				out.println("Start justification - Substituting hedge variables.");
			for (VariableWithHedges elem : uraucSys.getStoreHorizontalS()) {
				leftInstance.substitute(elem.getVar(), elem.getLeft());
				rightInstance.substitute(elem.getVar(), elem.getRight());
			}
			if (out != null && debugLevel == DebugLevel.PROGRESS) {
				out.println("Left  hand side - hedge variables instantiated: "
						+ leftInstance);
				out.println("Right hand side - hedge variables instantiated: "
						+ rightInstance);
				out.println("Substituting context variables.");
			}
			for (VariableWithHedges elem : uraucSys.getStoreVerticalQ()) {
				leftInstance.substitute(elem.getVar(), elem.getLeft());
				rightInstance.substitute(elem.getVar(), elem.getRight());
			}
			if (out != null && debugLevel == DebugLevel.PROGRESS) {
				out.println("Left  hand side - all variables instantiated: "
						+ leftInstance);
				out.println("Right hand side - all variables instantiated: "
						+ rightInstance);
			}
			if (!leftInstance.equals(leftHedge)
					|| !rightInstance.equals(rightHedge)) {
				if (out != null && debugLevel != DebugLevel.SILENT) {
					out.println("Back substitution: " + leftInstance + " =^= "
							+ rightInstance);
				}
				throw new JustificationException(
						"Justification of the computed anti-unifier failed!");
			}
			if (out != null && debugLevel != DebugLevel.SILENT)
				out.println("Justification ok.");
		}

		/**
		 * Returns the {@linkplain HedgeVar hedge variable} which is used as
		 * generalization variable. (The most general generalization.)
		 */
		public HedgeVar getMostGeneral() {
			return mostGeneral;
		}

		/**
		 * Returns the entire {@linkplain AntiUnifySystem} which is encapsulated
		 * by this class
		 */
		public AntiUnifySystem getResult() {
			return uraucSys;
		}
	}
}
