/*
 * Copyright 2013 Alexander Baumgartner
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package at.jku.risc.stout.urauc.algo;

import java.util.Arrays;
import java.util.Iterator;

import at.jku.risc.stout.urauc.algo.Contour.ContourAtom;
import at.jku.risc.stout.urauc.data.atom.TermAtom;

/**
 * A list of all the {@linkplain Contour}s of the common subsequence problem of
 * two lists of {@linkplain TermAtom}s. It also implements an iterator to
 * iterate all the subsets of {@linkplain Contour}s of a certain length, which
 * is needed if one wants to compute all the subsequences of a certain length.
 * 
 * @see AlignFnc
 * @author Alexander Baumgartner
 */
public class ContourArrangement implements Iterator<ContourArrangement>,
		Cloneable {
	private Contour[] contourList;
	private int size;
	// sub contour iterator
	private ContourArrangement subcontour;
	private boolean subcontourHasNext;
	private int subcontourChangeAt;
	private int[] subcontourIdx;

	public ContourArrangement() {
		this(16);
	}

	public ContourArrangement(int capacity) {
		contourList = new Contour[capacity];
	}

	/**
	 * Adds a new element to a certain {@linkplain Contour} and creates a new
	 * {@linkplain Contour} if necessary.
	 */
	public void setMatchAtom(int contour, int i, int j, TermAtom tLeft,
			TermAtom tRight) {
		if (size <= contour)
			size = contour + 1;
		if (contourList.length < size)
			contourList = Arrays.copyOf(contourList, size + (size >> 1));
		if (contourList[contour] == null)
			contourList[contour] = new Contour();
		contourList[contour].add(i, j, tLeft, tRight);
	}

	public int size() {
		return size;
	}

	public void clear() {
		for (int i = 0; i < size; i++)
			contourList[i].clear();
		size = 0;
	}

	/**
	 * Returns the {@linkplain Contour} at the given position.
	 */
	public Contour getContour(int contour) {
		return contourList[contour];
	}

	/**
	 * This is a shortcut for {@linkplain #getContour(int)}
	 * {@linkplain Contour#clip(ContourAtom) .clip(ContourElem)}.
	 */
	public Contour getClip(int contour, ContourAtom elem) {
		return contourList[contour].clip(elem);
	}

	public String toString() {
		StringBuilder sb = new StringBuilder();
		for (int i = size; --i >= 0;) {
			sb.append("Length ").append(i + 1).append(" contour:\n");
			sb.append(contourList[i]).append('\n');
		}
		return sb.toString();
	}

	/**
	 * Initialize the iteration of all the {@linkplain #size()} over k possible
	 * combinations of k contours.
	 */
	public void initIterator(int subcontourLen) {
		subcontour = new ContourArrangement(subcontourLen);
		subcontour.size = subcontourLen;
		subcontourHasNext = (subcontourLen <= size);
		if (subcontourHasNext) {
			System.arraycopy(contourList, 0, subcontour.contourList, 0,
					subcontourLen);
			subcontourIdx = new int[subcontourLen];
			subcontourChangeAt = 0;
			while (--subcontourLen >= 0)
				subcontourIdx[subcontourLen] = subcontourLen;
		}
	}

	@Override
	public boolean hasNext() {
		return subcontourHasNext;
	}

	@Override
	public ContourArrangement next() {
		int len = subcontour.size;
		// bulk copy from last changed position
		for (int j = subcontourChangeAt; j < len; j++)
			subcontour.contourList[j] = contourList[subcontourIdx[j]];

		int i = len - 1;
		// find the next possible contour to increment
		for (int end = size - len; i >= 0 && subcontourIdx[i] == end + i; i--)
			;
		if (i >= 0) {
			subcontourIdx[i]++;
			// pull back all the higher contours
			for (int start = subcontourIdx[i] - i, j = i + 1; j < len; j++)
				subcontourIdx[j] = start + j;
		} else {
			subcontourHasNext = false;
		}
		subcontourChangeAt = i;
		return subcontour;
	}

	@Override
	public void remove() {
		throw new UnsupportedOperationException("Remove is not supported.");
	}
}