/*
 * Copyright 2013 Alexander Baumgartner
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package at.jku.risc.stout.urauc.data;

import java.io.IOException;
import java.io.Writer;
import java.util.Collections;
import java.util.List;
import java.util.Map;
import java.util.Set;

import at.jku.risc.stout.urauc.data.atom.TermAtom;
import at.jku.risc.stout.urauc.data.atom.Variable;
import at.jku.risc.stout.urauc.util.DataStructureFactory;
import at.jku.risc.stout.urauc.util.IntList;
import at.jku.risc.stout.urauc.util.Printable;

/**
 * A {@linkplain Hedge} is a sequence of {@linkplain TermNode}s.
 * 
 * @author Alexander Baumgartner
 */
public class Hedge extends Printable implements Cloneable, Comparable<Hedge> {

	public static String PRINT_PARAM_START = "(";
	public static String PRINT_PARAM_SEPARATOR = ", ";
	public static String PRINT_PARAM_END = ")";
	private List<TermNode> sequence;
	private List<TermNode> sortedSequence;
	private boolean commutative;
	public static final Hedge nullHedge = new Hedge() {
		@Override
		public void add(TermNode t) {
			throw new IllegalAccessError("Illegal null-hedge modification");
		}

		@Override
		public void add(TermAtom atom) {
			throw new IllegalAccessError("Illegal null-hedge modification");
		}

		@Override
		public void setCommutative(boolean commutative) {
		}

		@Override
		public void makeCommutative(Set<TermAtom> commutativeSymbols,
				boolean commutative) {
		}

		@Override
		public Hedge clone() {
			return this;
		}
	};

	/**
	 * Create an empty hedge
	 */
	public Hedge() {
		this.sequence = DataStructureFactory.$.newList();
		this.sortedSequence = sequence;
	}

	/**
	 * Create an empty hedge
	 */
	public Hedge(boolean commutative) {
		this.sequence = DataStructureFactory.$.newList();
		this.sortedSequence = sequence;
		this.commutative = commutative;
	}

	/**
	 * Create an empty hedge with the given initial capacity.
	 */
	public Hedge(int capacity) {
		this.sequence = DataStructureFactory.$.newList(capacity);
		this.sortedSequence = sequence;
	}

	/**
	 * Create an empty hedge with the given initial capacity.
	 */
	public Hedge(int capacity, boolean commutative) {
		this.sequence = DataStructureFactory.$.newList(capacity);
		this.sortedSequence = sequence;
		this.commutative = commutative;
	}

	public TermAtom get(int[] position, int idx) {
		int termIdx = position[idx];
		if (termIdx >= size())
			return null;
		return get(termIdx).get(position, idx + 1);
	}

	/**
	 * Create a hedge with the specified sequence of {@linkplain TermNode}s in
	 * it. Changes in the hedge are reflected in this list, and vice-versa.
	 */
	public Hedge(List<TermNode> sequence) {
		this.sequence = sequence;
		this.sortedSequence = sequence;
	}

	public Hedge(List<TermNode> sequence, boolean commutative) {
		this.sequence = sequence;
		this.sortedSequence = sequence;
		this.commutative = commutative;
	}

	/**
	 * Create a hedge with the specified {@linkplain TermNode}s in it.
	 */
	public Hedge(TermNode... tNodes) {
		this(tNodes.length + 2);
		addAll(tNodes);
	}

	/**
	 * Create a hedge with {@linkplain TermNode}s which encapsulate the
	 * specified {@linkplain TermAtom}s in it.
	 */
	public Hedge(TermAtom... atoms) {
		this(atoms.length + 2);
		addAll(atoms);
	}

	public Hedge substitute(Variable x, TermNode t) {
		for (int i = sequence.size(); --i >= 0;)
			replace(i, sequence.get(i).substitute(x, t));
		return this;
	}

	public Hedge substitute(Variable x, Hedge h) {
		if (h.size() == 1)
			return substitute(x, h.get(0));
		return substitute(x, new TermNode(null, h));
	}

	public Hedge apply(Map<Variable, TermNode> sigma) {
		for (int i = sequence.size(); --i >= 0;)
			replace(i, sequence.get(i).apply(sigma));
		return this;
	}

	public boolean isEmpty() {
		return sequence.size() == 0;
	}

	protected List<TermNode> getSortedSequence() {
		if (commutative && sequence == sortedSequence && sequence.size() > 1) {
			sortedSequence = DataStructureFactory.$.newList(sequence);
			Collections.sort(sortedSequence);
		}
		return sortedSequence;
	}

	public void add(TermNode t) {
		if (t.isNullAtom()) {
			addAll(t.getHedge());
		} else {
			sequence.add(t);
			sortedSequence = sequence;
		}
	}

	public void add(TermAtom atom) {
		sequence.add(new TermNode(atom));
		sortedSequence = sequence;
	}

	public void addAll(TermNode... tNodes) {
		for (TermNode t : tNodes)
			add(t);
	}

	public void addAll(TermAtom... atoms) {
		for (TermAtom atom : atoms)
			add(atom);
	}

	public void addAll(Hedge other) {
		sequence.addAll(other.sequence);
		sortedSequence = sequence;
	}

	public TermNode get(int i) {
		return sequence.get(i);
	}

	public int size() {
		return sequence.size();
	}

	public boolean isCommutative() {
		return commutative;
	}

	public void setCommutative(boolean commutative) {
		this.commutative = commutative;
	}

	public boolean isHole() {
		return sequence.size() == 1 && sequence.get(0).isHole();
	}

	public TermAtomList top() {
		TermAtomList list = new TermAtomList(sequence.size());
		list.size = sequence.size();
		for (int i = 0, n = list.size; i < n; i++)
			list.atoms[i] = sequence.get(i).getAtom();
		return list;
	}

	/**
	 * Replaces the hole in this hedge with the given other hedge.Returns true
	 * if a hole has been found.
	 */
	public boolean replaceHole(Hedge plugIn) {
		for (int i = sequence.size(); --i >= 0;) {
			TermNode elem = sequence.get(i);
			if (elem.isHole()) {
				replace(i, plugIn);
				return true;
			}
			if (elem.getHedge().replaceHole(plugIn))
				return true;
		}
		return false;
	}

	/**
	 * Attention the index is stored in reverse order!
	 */
	public IntList getHoleIdxRec() {
		for (int i = sequence.size(); --i >= 0;) {
			TermNode node = sequence.get(i);
			if (node.isHole()) {
				IntList holeIdx = new IntList();
				holeIdx.add(i);
				return holeIdx;
			}
			IntList recIdx = node.getHedge().getHoleIdxRec();
			if (recIdx != null) {
				recIdx.add(i);
				return recIdx;
			}
		}
		return null;
	}

	public int getHoleIdx() {
		for (int i = sequence.size(); --i >= 0;) {
			TermNode node = sequence.get(i);
			if (node.isHole())
				return i;
			if (node.getHedge().getHoleIdx() > -1)
				return i;
		}
		return -1;
	}

	@Override
	public boolean equals(Object o) {
		if (this == o)
			return true;
		if (o instanceof Hedge)
			return getSortedSequence().equals(((Hedge) o).getSortedSequence());
		return true;
	}

	@Override
	public int hashCode() {
		return getSortedSequence().hashCode();
	}

	@Override
	public Hedge clone() {
		Hedge clone = new Hedge(sequence.size(), commutative);
		for (int i = 0, n = sequence.size(); i < n; i++)
			clone.sequence.add(sequence.get(i).clone());
		return clone;
	}

	@Override
	public int compareTo(Hedge other) {
		if (this == other)
			return 0;
		List<TermNode> sortedThis = this.getSortedSequence();
		List<TermNode> sortedOther = other.getSortedSequence();
		int len = sortedThis.size();
		if (len != sortedOther.size())
			return len - sortedOther.size();
		for (int i = 0; i < len; i++) {
			int cmp = sortedThis.get(i).compareTo(sortedOther.get(i));
			if (cmp != 0)
				return cmp;
		}
		return 0;
	}

	@Override
	public void print(Writer out) throws IOException {
		print(out, sequence.size() != 0);
	}

	public void print(Writer out, boolean parenthesis) throws IOException {
		if (parenthesis)
			out.append(PRINT_PARAM_START);
		for (int i = 0, n = sequence.size(); i < n; i++) {
			if (i > 0)
				out.append(PRINT_PARAM_SEPARATOR);
			sequence.get(i).print(out);
		}
		if (parenthesis)
			out.append(PRINT_PARAM_END);
	}

	/**
	 * Replaces the {@linkplain TermNode} at the given position with a new
	 * {@linkplain TermNode} if the {@linkplain TermAtom} of the new node is not
	 * null. Otherwise the old {@linkplain TermNode} will be replaced by the
	 * given sequence.
	 * 
	 * @return Returns the amount of nodes inserted instead of the old one.
	 */
	public int replace(int idx, TermNode toTerm) {
		if (toTerm.isNullAtom()) {
			return replace(idx, toTerm.getHedge());
		} else {
			sequence.set(idx, toTerm);
			sortedSequence = sequence;
			return 1;
		}
	}

	/**
	 * Attention: Breaks the position indexes!<br>
	 * Replaces the {@linkplain TermNode} at the given position by the given
	 * sequence.
	 * 
	 * @return Returns the amount of nodes inserted instead of the old one.
	 * @see #initPosition(int[])
	 */
	public int replace(int idx, Hedge toHedge) {
		// if (size() == 1)
		// commutative = toHedge.commutative;
		// else if (toHedge.size() > 1 && !toHedge.commutative)
		// commutative = false;
		sortedSequence = sequence;
		List<TermNode> oSequence = toHedge.sequence;
		int size = oSequence.size();
		switch (size) {
		case 0:
			sequence.remove(idx);
			return 0;
		case 1:
			sequence.set(idx, oSequence.get(0));
			return 1;
		default:
			sequence.set(idx, oSequence.get(0));
			sequence.addAll(idx + 1, oSequence.subList(1, size));
			return size;
		}
	}

	/**
	 * Returns the word representation of this hedge
	 */
	public TermAtomList word() {
		TermAtomList atomList = new TermAtomList();
		word(atomList);
		return atomList;
	}

	/**
	 * Adds the word representation of this hedge at the end of the list
	 */
	public void word(TermAtomList atomList) {
		for (int i = 0, n = sequence.size(); i < n; i++)
			sequence.get(i).word(atomList);
	}

	/**
	 * Returns a view of the tail of this hedge starting at the specified
	 * fromIndex, inclusive.
	 * 
	 * @see #subHedge(int, int)
	 */
	public Hedge subHedge(int fromIndex) {
		return subHedge(fromIndex, size());
	}

	/**
	 * Returns a view of the portion of this hedge between the specified
	 * fromIndex, inclusive, and toIndex, exclusive.
	 * 
	 * @see List#subList(int, int)
	 */
	public Hedge subHedge(int fromIndex, int toIndex) {
		return new Hedge(sequence.subList(fromIndex, toIndex), commutative);
	}

	public boolean nextCommutativeArrangement() {
		if (commutative) {
			if (nextCommutativeArrangementIntern())
				return true;
			sortSequence(false);
		}
		for (int i = sequence.size(); --i >= 0;)
			if (sequence.get(i).nextCommutativeArrangement())
				return true;
		return false;
	}

	private boolean nextCommutativeArrangementIntern() {
		int border = sequence.size() - 1;
		while (sequence.get(border).ignore && border > 0)
			border--;
		if (border < 1)
			return false;

		int leftIdx = border;
		TermNode rightNode, leftNode = sequence.get(leftIdx);
		do {
			if (leftIdx < 1)
				return false;
			rightNode = leftNode;
			leftNode = sequence.get(--leftIdx);
		} while (leftNode.compareTo(rightNode) > -1);

		int rightIdx = border;
		while (leftNode.compareTo(sequence.get(rightIdx)) > -1)
			rightIdx--;

		// Swap term nodes and correct swapped position indexes
		int idx = sequence.get(0).getAtom().getPositionIndex().length - 1;
		TermNode toSwap = sequence.set(leftIdx, sequence.get(rightIdx));
		toSwap.setPositionRecurse(idx, rightIdx);
		sequence.set(rightIdx, toSwap).setPositionRecurse(idx, leftIdx);
		leftIdx++;
		for (int i = 0, n = (border + 1 - leftIdx) / 2; i < n; i++) {
			int lTmp = leftIdx + i;
			int rTmp = border - i;
			toSwap = sequence.set(lTmp, sequence.get(rTmp));
			toSwap.setPositionRecurse(idx, rTmp);
			sequence.set(rTmp, toSwap).setPositionRecurse(idx, lTmp);
		}
		return true;
	}

	public void makeCommutative(Set<TermAtom> commutativeSymbols,
			boolean commutative) {
		this.commutative = commutative;
		if (!commutativeSymbols.isEmpty())
			for (int i = sequence.size(); --i >= 0;)
				sequence.get(i).makeCommutative(commutativeSymbols);
	}

	public void makeAssociative(Set<TermAtom> associativeSymbols,
			TermAtom associativeParent) {
		if (!associativeSymbols.isEmpty()) {
			int[] parentPos = null;
			for (int i = 0, n = sequence.size(); i < n;) {
				TermNode node = sequence.get(i);
				node.makeAssociative(associativeSymbols);
				if (node.getAtom().equals(associativeParent)) {
					replace(i, node.getHedge());
					n = sequence.size();
					parentPos = associativeParent.getPositionIndex();
				} else {
					if (parentPos != null)
						sequence.get(i).initPosition(parentPos, i);
					i++;
				}
			}
		}
	}

	public void sortCommutative(boolean sortAll) {
		for (int i = sequence.size(); --i >= 0;)
			sequence.get(i).sortCommutative(sortAll);
		if (commutative)
			sortSequence(sortAll);
	}

	protected void sortSequence(boolean sortAll) {
		int len = sequence.size();
		if (len < 2)
			return;
		List<TermNode> sortedSequence = getSortedSequence();
		// Sort and adjust positions
		for (int idx = sortedSequence.get(0).getAtom().getPositionIndex().length - 1, i = 0, sortIdx = 0; i < len; i++) {
			TermNode tNode = sortedSequence.get(i);
			if (sortAll || !tNode.ignore) {
				if (tNode != sequence.set(sortIdx, tNode))
					tNode.setPositionRecurse(idx, sortIdx);
				sortIdx++;
			}
		}
	}

	public void initPosition(int[] parentIndex) {
		for (int i = 0, n = size(); i < n; i++)
			sequence.get(i).initPosition(parentIndex, i);
	}

	public void normalizeVarNames(Map<String, String> renaming) {
		for (TermNode node : getSortedSequence())
			node.normalizeVarNames(renaming);
		sortedSequence = sequence;
	}

	public void setIgnoreLeaves(Map<TermAtom, Integer> symDiff) {
		boolean possiblyIgnore = size() - countIgnoreable(symDiff) < 2;
		for (int i = 0, n = size(); i < n;) {
			TermNode node = sequence.get(i);
			node.ignore = commutative
					&& (possiblyIgnore || Integer.valueOf(1).equals(
							symDiff.get(node.getAtom())))
					&& mayIgnore(node, symDiff);
			if (node.ignore) {
				int idx = node.getAtom().getPositionIndex().length - 1;
				node = sequence.set(--n, node);
				node.setPositionRecurse(idx, i);
				sequence.set(i, node);
			} else {
				++i;
				node.getHedge().setIgnoreLeaves(symDiff);
			}
		}
	}

	private int countIgnoreable(Map<TermAtom, Integer> symDiff) {
		int count = 0;
		if (commutative)
			for (TermNode node : sequence)
				if (mayIgnore(node, symDiff))
					count++;
		return count;
	}

	private boolean mayIgnore(TermNode node, Map<TermAtom, Integer> symDiff) {
		return node.getHedge().isEmpty() && symDiff.containsKey(node.getAtom());
	}

	public Hedge getHoleHedge() {
		for (TermNode node : sequence) {
			if (node.isHole())
				return this;
			Hedge hh = node.getHedge().getHoleHedge();
			if (hh != null)
				return hh;
		}
		return null;
	}

	public void setTo(TermNode singleNode) {
		sequence = DataStructureFactory.$.newList();
		sequence.add(singleNode);
		sortedSequence = sequence;
	}
}
