/*
 * Copyright 2013 Alexander Baumgartner
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package at.jku.risc.stout.urauc.data;

import java.util.Deque;

import at.jku.risc.stout.urauc.data.atom.ContextVar;
import at.jku.risc.stout.urauc.data.atom.FunctionSymbol;
import at.jku.risc.stout.urauc.data.atom.FunctionVar;
import at.jku.risc.stout.urauc.data.atom.HedgeVar;
import at.jku.risc.stout.urauc.data.atom.Hole;
import at.jku.risc.stout.urauc.data.atom.TermAtom;
import at.jku.risc.stout.urauc.data.atom.TermVar;
import at.jku.risc.stout.urauc.data.atom.Variable;
import at.jku.risc.stout.urauc.util.DataStructureFactory;

/**
 * A factory to create {@linkplain TermNode}s, {@linkplain Hedge}s and atomic
 * types ( {@linkplain TermAtom}s).
 * 
 * @author Alexander Baumgartner
 */
public class NodeFactory {

	public static String PREFIX_ContextVar;
	public static String PREFIX_FunctionVar;
	public static String PREFIX_HedgeVar;
	public static String PREFIX_IndividualVar;
	public static String PREFIX_Function;
	public static String PREFIX_Constant;
	public static String PREFIX_FreshHedgeVar = "#";
	public static String SUFFIX_FreshHedgeVar = "";
	public static String PREFIX_FreshCtxVar = "$";
	public static String SUFFIX_FreshCtxVar = "";
	public static String PREFIX_FreshTermVar = "!";
	public static String SUFFIX_FreshTermVar = "";
	public static String PREFIX_FreshFncVar = "&";
	public static String SUFFIX_FreshFncVar = "";

	public static long hedgeVarCnt = 0;
	public static long ctxVarCnt = 0;

	private Deque<Hedge> hStack = DataStructureFactory.$.newDeque();
	private static final TermNode hole = new TermNode(new Hole("@"), null);

	public static void resetCounter() {
		hedgeVarCnt = 0;
		ctxVarCnt = 0;
	}

	public void pushHedge() {
		hStack.push(new Hedge());
	}

	public void addToHedge(TermNode node) {
		hStack.peek().add(node);
	}

	public Hedge popHedge() {
		Hedge hRet = hStack.pop();
		if (hRet.size() == 0)
			return Hedge.nullHedge;
		return hRet;
	}

	public Hedge createHedge(TermNode... nodes) {
		if (nodes.length == 0)
			return Hedge.nullHedge;
		Hedge h = new Hedge();
		for (TermNode termNode : nodes) {
			h.add(termNode);
		}
		return h;
	}

	public TermNode createContextVar(String name, Hedge hedge) {
		if (PREFIX_ContextVar != null)
			name = PREFIX_ContextVar + name;
		return new TermNode(new ContextVar(name), hedge);
	}

	public TermNode createFunctionVar(String name, Hedge hedge) {
		if (PREFIX_FunctionVar != null)
			name = PREFIX_FunctionVar + name;
		return new TermNode(new FunctionVar(name), hedge);
	}

	public TermNode createHedgeVar(String name) {
		if (PREFIX_HedgeVar != null)
			name = PREFIX_HedgeVar + name;
		return new TermNode(new HedgeVar(name), null);
	}

	public TermNode createIndividualVar(String name) {
		if (PREFIX_IndividualVar != null)
			name = PREFIX_IndividualVar + name;
		return new TermNode(new TermVar(name), null);
	}

	public TermNode createFunction(String name, Hedge hedge) {
		if (PREFIX_Function != null)
			name = PREFIX_Function + name;
		return new TermNode(new FunctionSymbol(name), hedge);
	}

	public TermNode createConstant(String name) {
		if (PREFIX_Constant != null)
			name = PREFIX_Constant + name;
		return new TermNode(new FunctionSymbol(name), null);
	}

	/**
	 * Will return always the same Object.
	 */
	public TermNode createHole(String name) {
		return hole;
	}

	public static HedgeVar obtainFreeHedgeVar() {
		StringBuilder name = new StringBuilder(PREFIX_FreshHedgeVar);
		name.append(++hedgeVarCnt);
		name.append(SUFFIX_FreshHedgeVar);
		return new HedgeVar(name.toString()) {
			@Override
			protected String getCompareName() {
				return "";
			}
		};
	}

	public static Variable obtainCorrespondingCtxVar(String varName) {
		return new ContextVar(obtainCtxVarName(varName)) {
			@Override
			protected String getCompareName() {
				return "";
			}
		};
	}

	public static String obtainCtxVarName(String varName) {
		StringBuilder name = new StringBuilder(PREFIX_FreshCtxVar);
		name.append(varName.subSequence(PREFIX_FreshHedgeVar.length(),
				varName.length() - SUFFIX_FreshHedgeVar.length()));
		name.append(SUFFIX_FreshCtxVar);
		return name.toString();
	}

	public static String obtainHedgeVarName(String varName) {
		StringBuilder name = new StringBuilder(PREFIX_FreshHedgeVar);
		name.append(varName.subSequence(PREFIX_FreshCtxVar.length(),
				varName.length() - SUFFIX_FreshCtxVar.length()));
		name.append(SUFFIX_FreshHedgeVar);
		return name.toString();
	}

	public static TermNode obtainFreeHedgeNode() {
		return toHedgeNode(obtainFreeHedgeVar());
	}

	public static TermNode toHedgeNode(HedgeVar hedgeVar) {
		return newNode(hedgeVar, null);
	}

	public static TermNode newNode(TermAtom atom, Hedge hedge) {
		return new TermNode(atom, hedge);
	}

	public static TermNode obtainHole() {
		return hole;
	}
}
