/*
 * Copyright 2013 Alexander Baumgartner
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package at.jku.risc.stout.urauc.data.atom;

import static java.lang.Math.min;

import java.io.IOException;
import java.io.Writer;
import java.util.Arrays;
import java.util.Map;

import at.jku.risc.stout.urauc.data.TermNode;
import at.jku.risc.stout.urauc.util.Printable;

/**
 * Base class for all the atomic types like function symbols and variables.
 * 
 * @author Alexander Baumgartner
 */
public abstract class TermAtom extends Printable implements
		Comparable<TermAtom>, Cloneable {
	private String name;
	private boolean commutative;
	private int[] positionIndex;
	private Class<? extends TermAtom> clazz = getClass();

	// Ancestor test speedup
	// private int positionHash;
	// private int[] positionHashArray;
	private int[] cachePos;
	private int cacheVal;
	public static final TermAtom nullAtom = new TermAtom("",
			TermNode.POSITION_E) {
		@Override
		public void print(Writer out) throws IOException {
		}

		@Override
		public void setPositionIndex(int[] positionIndex) {
		}

		@Override
		public TermAtom clone() {
			return this;
		}
	};

	public static boolean DEBUG_POSITION = false;

	public TermAtom(String name) {
		setName(name);
	}

	protected void setName(String name) {
		this.name = name.intern();
	}

	public TermAtom(String name, int[] positionIndex) {
		this(name);
		this.positionIndex = positionIndex;
	}

	public int[] getPositionIndex() {
		return positionIndex;
	}

	public void setPositionIndex(int[] positionIndex) {
		this.positionIndex = positionIndex;
		// increase speed of isAncestor by using a position hash code.
		/*
		 * positionHashArray = new int[positionIndex.length]; int result = 1;
		 * for (int i = 0, n = positionIndex.length; i < n; i++) { result = 31 *
		 * result + positionIndex[i]; positionHashArray[i] = result; }
		 * positionHash = result;
		 */
	}

	public String getName() {
		return name;
	}

	@Override
	public boolean equals(Object obj) {
		if (obj instanceof TermAtom) {
			TermAtom atom = (TermAtom) obj;
			// We only need to compare the pointers, because of calling intern
			// at construction time
			return name == atom.name && clazz == atom.clazz;
		}
		return false;
	}

	@Override
	public TermAtom clone() {
		try {
			TermAtom clone = (TermAtom) super.clone();
			if (positionIndex != null)
				clone.positionIndex = Arrays.copyOf(positionIndex, positionIndex.length);
			return clone;
		} catch (CloneNotSupportedException e) {
			throw new RuntimeException(e);
		}
	}

	@Override
	public int hashCode() {
		return name.hashCode();
	}

	@Override
	public int compareTo(TermAtom o) {
		String n1 = getCompareName(), n2 = o.getCompareName();
		if (n1 != n2)
			return n1.compareTo(n2);
		Class<? extends TermAtom> c1 = clazz;
		Class<? extends TermAtom> c2 = o.clazz;
		if (c1 == c2)
			return 0;
		return c1.getSimpleName().compareTo(c2.getSimpleName());
	}

	protected String getCompareName() {
		return name;
	}

	@Override
	public void print(Writer out) throws IOException {
		out.append(getName());
		if (positionIndex != null && DEBUG_POSITION)
			out.append(Arrays.toString(positionIndex));
	}

	public TermNode substitute(Variable from, TermNode to, TermNode thisNode) {
		return thisNode;
	}

	public TermNode apply(Map<Variable, TermNode> sigma, TermNode thisNode) {
		return thisNode;
	}

	public boolean isAncestor(TermAtom other) {
		int n = positionIndex.length;
		// Fast test
		if (n >= other.positionIndex.length)
			// || positionHash != other.positionHashArray[n - 1])
			return false;
		// Exact test
		return n == commonAncestor(other);
	}

	public int commonAncestor(TermAtom other) {
		// Same int[] instance as before?
		if (this.cachePos == other.positionIndex)
			return this.cacheVal;
		if (other.cachePos == this.positionIndex)
			return other.cacheVal;
		cachePos = other.positionIndex;
		cacheVal = 0;
		for (int n = min(positionIndex.length, cachePos.length);; cacheVal++)
			if (cacheVal == n || positionIndex[cacheVal] != cachePos[cacheVal])
				return cacheVal;
	}

	public String getPositionString(int idxStart, int idx1Add) {
		StringBuilder sb = new StringBuilder(positionIndex.length * 3);
		for (int i = idxStart, n = positionIndex.length; i < n; i++) {
			int idx = positionIndex[i] + 1; // Start with 1 instead of 0
			if (i == idxStart)
				sb.append(idx + idx1Add);
			else
				sb.append('.').append(idx);
		}
		return sb.toString();
	}

	public boolean isCommutative() {
		return commutative;
	}

	public void setCommutative(boolean commutative) {
		this.commutative = commutative;
	}

	public void normalizeVarNames(Map<String, String> renaming) {
	}
}
