/*
 * Copyright 2013 Alexander Baumgartner
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

/*
 * Copyright 2012 Alexander Baumgartner
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package at.jku.risc.stout.urauc.util;

/**
 * A tiny abstract class for list implementations. All implementors have to
 * provide O(1) complexity for all the abstract methods!!! Every method call
 * other then {@linkplain #toString()} which is defined in this class has
 * complexity O(1).
 * 
 * @author Alexander Baumgartner
 */
public abstract class TinyList<V> implements Cloneable {
	protected int start;
	protected int size;

	/**
	 * The relative size of this list is defined by {@linkplain #size} -
	 * {@linkplain #start}.
	 */
	public int size() {
		return size - start;
	}

	/**
	 * Returns the element at the absolute position.
	 * 
	 * @see #get(int)
	 */
	protected abstract V getAbsolut(int i);

	/**
	 * Returns the element at the relative position i. The relative position is
	 * defined by {@linkplain #getAbsolut(int)} with argument
	 * {@linkplain #start} + i.
	 */
	public V get(int i) {
		return getAbsolut(start + i);
	}

	/**
	 * Removes the last element by decrementing the size.<br>
	 * This is a dirty remove which does not nullify any pointers!
	 */
	public void removeLast() {
		size--;
	}

	/**
	 * Removes the first element by incrementing the start.<br>
	 * This is a dirty remove which does not nullify any pointers!
	 */
	public void removeFirst() {
		this.start++;
	}

	/**
	 * Returns the last element.
	 */
	public V getLast() {
		return getAbsolut(size - 1);
	}

	/**
	 * Returns the (relative) first element.
	 */
	public V getFirst() {
		return getAbsolut(start);
	}

	/**
	 * Returns true if the relative size is 0.
	 */
	public boolean isEmpty() {
		return size() == 0;
	}

	public <T extends TinyList<V>> T subList(int from, int to) {
		try {
			@SuppressWarnings("unchecked")
			T ret = (T) super.clone();
			ret.start += from;
			ret.size = start + to;
			return ret;
		} catch (CloneNotSupportedException e) {
			throw new RuntimeException(e);
		}
	}

	public <T extends TinyList<V>> T subList(int from) {
		try {
			@SuppressWarnings("unchecked")
			T ret = (T) super.clone();
			ret.start += from;
			return ret;
		} catch (CloneNotSupportedException e) {
			throw new RuntimeException(e);
		}
	}

	@Override
	public String toString() {
		StringBuilder sb = new StringBuilder();
		sb.append('[');
		if (size > start) {
			sb.append(getAbsolut(start));
			for (int i = start; ++i < size;)
				sb.append(',').append(getAbsolut(i));
		}
		sb.append(']');
		return sb.toString();
	}
}
